package com.anteatersecurity.test;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.security.GeneralSecurityException;
import java.security.Provider;
import java.security.SecureRandom;
import java.security.Security;

import alt.bouncycastle.jce.provider.BouncyCastleProvider;


import es.vocali.util.AESCrypt;


public class BouncyCastleAESCryptMemoryLeakExample {
	
	final static int ONE_KILOBYTE = 1024;
	final static int ONE_MEGABYTE = ONE_KILOBYTE*1024;
	final static String PASSWORD = "We need a password, but we don't care what the value is, so we just use this static value. We don't care what the value is, because 1) we don't need to decrypt and 2) there isn't any sensitive data to protect.";
	final static int NUMBER_OF_ENCRYPTIONS = 50;
	
	
	public static void main(String args []) throws Exception {
		runTests();
	}

	public static void runTests() {
//		Uncomment the next two lines of code to switch BouncyCastle provider 
//		from org.bouncycastle to alt.bouncycastle.
//		Security.removeProvider(BouncyCastleProvider.PROVIDER_NAME);
//		Security.addProvider(new BouncyCastleProvider());
		Thread thread = new Thread(new Runnable() {

			@Override
			public void run() {
				Provider providers[] = Security.getProviders();
				for(int i = 0; i < providers.length; i++) {
					System.out.println("provider[" + i + "] = " + providers[i].getName());
				}
				System.out.println("BouncyCastle Crypto Provider class: "
						+ Security.getProvider(BouncyCastleProvider.PROVIDER_NAME)
								.getClass().getName());
				BouncyCastleAESCryptMemoryLeakExample poc = new BouncyCastleAESCryptMemoryLeakExample();
				
				for (int i = 0; i < 10; i++) {
					try {
						poc.runTest();
						System.out.println();
						System.out.println();
						System.out.println();
						System.out.println();
						System.out.println();
					} catch (IOException e) {
						e.printStackTrace();
					} catch (GeneralSecurityException e) {
						e.printStackTrace();
					}	
				}
			}
		});
		thread.start();
	}

	private void runTest() throws IOException, GeneralSecurityException {
		System.out
				.println("This test encrypts a megabyte of data "
						+ NUMBER_OF_ENCRYPTIONS
						+ " times then displays the memory usage. "); 
		System.out.println("Only a ByteArrayInputStream object is created for " +
						"each iteration.");
		System.out.println("The ByteArrayInputStream's data, the " +
						"OutputStream and the AESCrypt objects are all reused, ");
		System.out.println("so the memory usage should be nearly constant. If " +
						"memory use comes anywhere near 1 megabyte per ");
		System.out.println("iteration, then there is a leak.");
		
		byte[] data = getData(ONE_MEGABYTE);
		
		//The megabyte allows room for the encrypted data. The kilobyte allows
		//room for the initialization vector, padding and anything else that
		//gets appended during encryption. Note that AES modifies and shifts
		//bits, it does not encode or otherwise expand the individual bytes of 
		//data. 
		ByteArrayOutputStream baos = new ByteArrayOutputStream(ONE_MEGABYTE+ONE_KILOBYTE);
		
		//We reuse the password, because this test is only concerned with 
		//memory allocation -not security.
		AESCrypt aesCrypt = new AESCrypt(PASSWORD);
		
		//This provides a baseline before we start our tests.
		forceGarbageCollection();
		System.out.println("Before starting encryption.");
		long startingFreeMemory = Runtime.getRuntime().freeMemory();
		printMemoryStatistics();
		System.out.println("");
		System.out.println("Starting encryption. Memory statistics will display after each iteration");
		System.out.println("------------------------------------------------------------------------");
		for(int i = 0; i < NUMBER_OF_ENCRYPTIONS; i++) {
			System.out.print("Iteration " + i + "\t");
			aesEncrypt(aesCrypt, data, baos);
		}
		System.out.println("------------------------------------------------------------------------");
		System.out.println("Encryption complete.");
		long endingFreeMemory = Runtime.getRuntime().freeMemory();
		
		forceGarbageCollection();
		long freeMemoryAfterGC = Runtime.getRuntime().freeMemory();
		System.out.println("Starting free memory: " + startingFreeMemory
				+ "\tEnding free memory " + endingFreeMemory
				+ "\tFree memory after GC: " + freeMemoryAfterGC);
		System.out
				.println("Megabyte difference between starting free memory and ending memory: "
						+ ((startingFreeMemory - endingFreeMemory)/ONE_MEGABYTE) + "MB");
		System.out
		.println("Megabyte difference between starting free memory and free memory after GC: "
				+ ((startingFreeMemory - freeMemoryAfterGC)/ONE_MEGABYTE) + "MB");
	}
	
	private void forceGarbageCollection() {
		Thread.yield();
		System.gc();
		Thread.yield();
		System.gc();
		Thread.yield();
		System.gc();
		Thread.yield();
	}

	private void printMemoryStatistics() {
		long totalMemory = Runtime.getRuntime().totalMemory();
		long freeMemory = Runtime.getRuntime().freeMemory();
		System.out.println("Total memory: " + totalMemory +// 
				"\t\tFree memory: " + freeMemory +//
				"\t\tMemory used (MB): " + ((totalMemory-freeMemory)/ONE_MEGABYTE) + "MB");
	}
	
	/**
	 * Return a byte array filled with arbitrary data. It doesn't matter what
	 * the data is. All that we care about is the size.
	 * @param size how large the byte array should be.
	 * @return A byte array of size <code>size</code>
	 */
	private byte[] getData(int size) {
		byte data[] = new byte[size];
		SecureRandom secureRandom = new SecureRandom();
		secureRandom.nextBytes(data);
		return data;
	}
	
	private void aesEncrypt(AESCrypt aesCrypt, byte[] data, ByteArrayOutputStream baos)
			throws IOException, GeneralSecurityException {
		//We create a new ByteArrayInputStream with an existing data array
		//All it does is assign the buffer, and set two indexes.
		ByteArrayInputStream bais = new ByteArrayInputStream(data);
		aesCrypt.encrypt(2, bais, baos);
//		If you want to run the test without calling AESCrypt in the line above,
//		Uncomment this line to give yourself enough time to view the App Manager
//		try {
//			Thread.sleep(250);
//		} catch (InterruptedException e) {
//			e.printStackTrace();
//		}
		
		//Neither the close, nor the null are strictly necessary
		bais.close();
		bais = null;
		//baos.reset() sets ByteArrayOutputStrem's this.count = 0 so that the 
		//original array can be reused. It does not allocate a new byte[].
		baos.reset();
		printMemoryStatistics();
	}
}